""" Discontinuous Data-Oriented Parsing (disco-dop).
Main components:

- A parser for Probalistic Linear Context-Free Rewriting Systems (LCFRS),
  as well as Probabilistic Context-Free Grammars (PCFG).
- Facilities to extract and parse with data-oriented parsing (DOP) grammars.
"""
