<!-- id={{ randid }} -->
<pre style="display: block;" id=n{{ randid }}>
{{ nbest }}
</pre>
<pre style="display: block;" id=f{{ randid }}>
{{ frags }}
</pre>
<pre style="display: block;" id=i{{ randid }}>
{{ info }}
</pre>
<pre>
{{ result }}
(<a href='javascript: togglelink("f{{ randid }}")' id=linkf{{ randid }}>hide fragments</a>; <a href='javascript: togglelink("n{{ randid }}")' id=linkn{{ randid }}>hide alternative analyses</a>; <a href='javascript: togglelink("i{{ randid }}")' id=linki{{ randid }}>hide info</a>; <a href="/?{{ link }}">link</a>)
</pre>
